
#ifndef IAITO_BINDINGS_H
#define IAITO_BINDINGS_H

// #define QT_ANNOTATE_ACCESS_SPECIFIER(a) __attribute__((annotate(#a)))

#include "../common/Configuration.h"
#include "../core/Iaito.h"
#include "../core/MainWindow.h"
#include "../menus/AddressableItemContextMenu.h"
#include "../plugins/IaitoPlugin.h"
#include "../widgets/IaitoDockWidget.h"

#endif // IAITO_BINDINGS_H
